/*
 * This file is part of LibEuFin.
 * Copyright (C) 2023-2024 Taler Systems S.A.

 * LibEuFin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation; either version 3, or
 * (at your option) any later version.

 * LibEuFin is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
 * Public License for more details.

 * You should have received a copy of the GNU Affero General Public
 * License along with LibEuFin; see the file COPYING.  If not, see
 * <http://www.gnu.org/licenses/>
 */
package tech.libeufin.bank.api

import io.ktor.http.*
import io.ktor.server.application.*
import io.ktor.server.response.*
import io.ktor.server.routing.*
import io.github.smiley4.ktoropenapi.*
import tech.libeufin.bank.BankConfig
import tech.libeufin.bank.TokenLogicalScope
import tech.libeufin.bank.auth.auth
import tech.libeufin.bank.bankInfo
import tech.libeufin.bank.db.Database
import tech.libeufin.common.HistoryParams
import tech.libeufin.common.RevenueConfig
import tech.libeufin.common.RevenueIncomingHistory

fun Routing.revenueApi(db: Database, cfg: BankConfig) { 
    get("/accounts/{USERNAME}/taler-revenue/config", {
        operationId = "getRevenueConfig"
        description = "Get revenue API configuration"
        tags = listOf("Revenue")
        request {
            pathParameter<String>("USERNAME") { description = "Account username" }
        }
        response {
            code(HttpStatusCode.OK) { description = "Revenue API configuration"; body<RevenueConfig>() }
        }
    }) {
        call.respond(RevenueConfig(
            currency = cfg.regionalCurrency
        ))
    }
    auth(db, cfg.pwCrypto, TokenLogicalScope.revenue, cfg.basicAuthCompat) {
        get("/accounts/{USERNAME}/taler-revenue/history", {
            operationId = "getRevenueHistory"
            description = "Get revenue history for an account"
            tags = listOf("Revenue")
            protected = true
            securitySchemeNames("bearerAuth", "basicAuth")
            request {
                pathParameter<String>("USERNAME") { description = "Account username" }
                queryParameter<Int>("limit") {
                    description = "Maximum number of results to return (default: -20, negative means descending)"
                    required = false
                }
                queryParameter<Long>("offset") {
                    description = "Row ID offset for pagination"
                    required = false
                }
                queryParameter<Long>("timeout_ms") {
                    description = "Timeout for long polling in milliseconds (default: 0, no long polling)"
                    required = false
                }
            }
            response {
                code(HttpStatusCode.OK) { description = "Revenue history"; body<RevenueIncomingHistory>() }
                code(HttpStatusCode.NoContent) { description = "No revenue history found" }
            }
        }) {
            val params = HistoryParams.extract(call.request.queryParameters)
            val bankAccount = call.bankInfo(db)
            val items = db.transaction.revenueHistory(params, bankAccount.bankAccountId)

            if (items.isEmpty()) {
                call.respond(HttpStatusCode.NoContent)
            } else {
                call.respond(RevenueIncomingHistory(items, bankAccount.payto))
            }
        }
    }
}